//	BPLUS.C
//
//	Version: 1.3a
//
//	---------------------------------------------------------------------------
//	BPLUS file indexing routines
//			by
//	Youri Tseytins.
//	---------------------------------------------------------------------------
//
//	Copyright (c) 2005
//
//  This software is provided 'as-is', without any express or implied
//  warranty.  In no event will the authors be held liable for any damages
//  arising from the use of this software.
//
//  Permission is granted to anyone to use this software for any purpose,
//  including commercial applications, and to alter it and redistribute it
//  freely, subject to the following restrictions:
//
//  1. The origin of this software must not be misrepresented; you must not
//     claim that you wrote the original software. If you use this software
//     in a product, an acknowledgment in the product documentation would be
//     appreciated but is not required.
//  2. Altered source versions must be plainly marked as such, and must not be
//     misrepresented as being the original software.
//  3. This notice may not be removed or altered from any source distribution.
//
//

// Sources changed to make it work on Opticon Terminals
// Martin Jansen


//#include "stdafx.h"

#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <unistd.h>
#if OPH1004 || OPN2002
#include <fcntl.h>
#include <sys\types.h>            // delete this line for Turbo C
#include <sys\stat.h>
#endif

#include <string.h>
#include <setjmp.h>

#include "bplus.h"

#include "Database.h"
#if defined(WIN32)
	#include <io.h>
#endif
#include "lib.h"

#if OPH1004 || OPN2002
#include <unistd.h>
#undef O_BINARY
#define O_BINARY 0x00
#endif

#ifdef WIN32
	#define open		_open
	#define close		_close
	#define read		_read
	#define write		_write
	#define lseek		_lseek
	#define O_RDWR		_O_RDWR
	#define O_BINARY	_O_BINARY
	#define O_CREAT		_O_CREAT
	#define O_TRUNC		_O_TRUNC
#endif

//
// macros, constants, data types
//
#define  NULLREC      (-1L)
#define  FREE_BLOCK   (-2)


#define  ENT_ADR(pb,off)  ((ENTRY*)(((pb)->entries) + off ))
#define	 ENT_SIZE		  sizeof(ENTRY)

#define  BUFDIRTY(j)      (mci.cache[j].dirty)
#define  BUFHANDLE(j)     (mci.cache[j].handle)
#define  BUFBLOCK(j)      (mci.cache[j].mb)
#define  BUFCOUNT(j)      (mci.cache[j].count)
#define  CB(j)            (pci->pos[j].cblock)
#define  CO(j)            (pci->pos[j].coffset)
//#define  TRUE             1
//#define  FALSE            0

// BPLUS uses the library routine
// memmove which must be used with
// Turboc 1.5, 2.0, Quick C and
// MS C 5.0 and 5.1
// #define memmove    memcpy      Use this macro for Microsoft C4.0

//
//  declare some global variables
//
jmp_buf	catchIndexError;	// Error handler

IX_DESC      *pci;
IX_BUFFER    mci;
BLOCK        *block_ptr;
BLOCK        *spare_block;
int          cache_ptr = 0;
int   		 cache_init = 0;
int          split_size = IXB_SPACE;
int          comb_size = (IXB_SPACE >> 1);

void  Error( int error, long recPtr );
void  ReadIf( long startPos, char *buf, int readBytes );
void  WriteIf(int fd, long startPos, char *buf, int writeBytes );
int   CreatIf( const char *idxName );
int	  OpenIf( const char *idxName );
void  CloseIf( int fd );

void ResetBuffers( IX_DESC *pix );
void UpdateBlock( void );
void InitCache( void );
int  FindCache( RECPOS r );
int  NewCache( void );
void LoadCache( RECPOS r );
void GetCache( RECPOS r );
void RetrieveBlock( int j, RECPOS r);
int  PrevEntry( int off);
int  NextEntry( int off );
void CopyEntry( ENTRY *to, ENTRY *from);
int  ScanBlock( int n );
int  LastEntry( void );
void WriteFree( RECPOS r, BLOCK *pb );
RECPOS GetFree( void );
int  FindBlock( ENTRY *pe, int *off, int *poff);
void MoveDown( BLOCK *pb, int off, int n);
void MoveUp( BLOCK *pb, int off, int n);
void InsertBlock( BLOCK *pb, ENTRY *pe, int off );
void DeleteBlock( BLOCK *pb, int off );
void Split( int l, ENTRY *pe, ENTRY *e );
void InsertLevel( int l, ENTRY *pe);
int  InsertIx( ENTRY *pe, IX_DESC *pix );
int  FindIx( ENTRY *pe, IX_DESC *pix, int find );
int  CombineBlock( RECPOS ads, int size );
void ReplaceEntry( ENTRY *pe );


//void BplusInit(void)
//{
//    /*
//    printf("BplusInit()   mci malloc: %ld bytes\r\n", sizeof(IX_BUFFER));
//
//    mci = (IX_BUFFER*) malloc (sizeof(IX_BUFFER));
//    if (mci==NULL)
//    {
//    }*/
//
//	cache_ptr = 0;
//	cache_init = 0;
//	split_size = IXB_SPACE;
//	comb_size  = (IXB_SPACE >> 1);
//}



//void print_blk2(BLOCK2 *);

//
// file I/O for BPLUS indexing routines
//

// TODO: change this later
void  Error(int error, long recPtr)
{
	// remove the message in the real version
    //static char *msg[3] = {"ERROR - CANNOT OPEN/CLOSE FILE",
    //                       "ERROR WHILE READING FILE",
    //                       "ERROR WHILE WRITING FILE"};
    //printf("\n  %s - Record Number %ld\n", msg[ error ], recPtr);
#ifdef WIN32
	if( error == 1 )
		longjmp( catchIndexError, ERR_INDEX_READ );
	else
		longjmp( catchIndexError, ERR_INDEX_WRITE );
//	exit(error);
#else
	if( error == 1 )
		longjmp( catchIndexError, ERR_INDEX_READ );
	else
		longjmp( catchIndexError, ERR_INDEX_WRITE );
//	for(;;)
//		delay(100);
#endif
}


void  ReadIf( long startPos, char *buf, int readBytes )
{
//#ifdef WIN32
//	if( _lseek( pci->ixfile, startPos, SEEK_SET ) == startPos )
//		if( _read( pci->ixfile, buf, readBytes ) == readBytes )
//			return;
//#else
	if( lseek( pci->ixfile, startPos, SEEK_SET ) == startPos )
		if( read( pci->ixfile, buf, readBytes ) == readBytes )
			return;
//#endif
	Error( 1, startPos );

}


void  WriteIf(int fd, long startPos, char *buf, int writeBytes)
{
//#ifdef WIN32
//	if( _lseek( fd, startPos, SEEK_SET ) == startPos )
//		if( _write( fd, buf, writeBytes ) == writeBytes )
//			return;
//#else

	if( lseek( fd, startPos, SEEK_SET ) == startPos )
		if( write( fd, buf, writeBytes ) == writeBytes )
			return;
//#endif
	Error( 2, startPos );
}


int CreatIf( const char *idxName )
{
    int	ret;
    ret = open( idxName, O_RDWR | O_CREAT | O_TRUNC | O_BINARY, 0777);
    return (ret);
}


int OpenIf( const char *idxName)
{
	int ret;
	ret = open( idxName, O_RDWR | O_APPEND | O_BINARY, 0777 );
    return (ret);
}

void CloseIf( int fd )
{
	if( close( fd ) < 0)
	{
		Error( 2, 0L);
	}
}


int OpenIndex( const char *idxName,  IX_DESC *pix, int dup)
{
	pci = pix;
    pci->ixfile = OpenIf( idxName );
	if( pci->ixfile < 0 )
		return (IX_FAIL);
    pci->root_dirty = FALSE;
    pci->duplicate = dup;
    ReadIf( 0L, (char *)&(pix->root), (sizeof(BLOCK) + sizeof(IX_DISK)));
    if( !cache_init )
    {
        InitCache();
        cache_init = 1;
	}
    return (IX_OK);
  }


void BufferFlush( IX_DESC *pix )
{
	int i;
	if( pix->root_dirty )
	{
		WriteIf( pix->ixfile, 0L,(char *)&(pix->root), (sizeof(BLOCK) + sizeof(IX_DISK)));
		pix->root_dirty = FALSE;
	}
	for( i = 0; i < NUM_BUFS; i++)
	{
		if( (BUFHANDLE(i) == pix->ixfile) && BUFDIRTY(i) )
		{
			WriteIf(BUFHANDLE(i), BUFBLOCK(i).brec, (char *) &BUFBLOCK(i), sizeof(BLOCK));
			BUFDIRTY(i) = FALSE;
		}
	}
}

void ResetBuffers( IX_DESC *pix )
{
	int i;
	for( i = 0; i < NUM_BUFS; i++)
		if( BUFHANDLE(i) == pix->ixfile)
			BUFBLOCK(i).brec = NULLREC;
}


int CloseIndex( IX_DESC *pix)
{
	int ret = IX_OK;

    BufferFlush( pix );
    ResetBuffers( pix );
    CloseIf( pix->ixfile );
    return( ret );
}

int MakeIndex( const char *idxName, IX_DESC *pix, int dup)
{
	pci = pix;
    pci->ixfile = CreatIf( idxName );
	if( pci->ixfile < 0 )
		return (IX_FAIL);
    pci->duplicate = dup;
    pci->root_dirty = FALSE;
    pci->dx.nl = 1;
    pci->dx.ff = NULLREC;
	pci->dx.totalkeys = 0;	// TEST TEST TEST
    pci->level = 0;
    CO(0) = -1;
    CB(0) = 0L;
    pci->root.brec = 0L;
    pci->root.bend = 0;
    pci->root.p0 = NULLREC;
    WriteIf( pci->ixfile, 0L,(char *)&(pix->root), (sizeof(BLOCK) + sizeof(IX_DISK)));
    if( !cache_init)
    {
        InitCache();
        cache_init = 1;
    }
    return ( IX_OK );
  }



//
// cache I/O for BPLUS indexing routines
//

void  UpdateBlock( void )
{
	if( block_ptr == &(pci->root))
		pci->root_dirty = TRUE;
    else
		BUFDIRTY( cache_ptr ) = TRUE;
}


void  InitCache( void )
{
	register int  j;
    for (j = 0; j < NUM_BUFS; j++)
    {
		BUFDIRTY(j) = 0;
		BUFCOUNT(j) = 0;
		BUFBLOCK(j).brec = NULLREC;
	}
}


int FindCache( RECPOS r)
{
	register int  j;
	for( j = 0; j < NUM_BUFS; j++)
	{
		if((BUFBLOCK(j).brec == r) && (BUFHANDLE(j) == pci->ixfile))
		{
			cache_ptr = j;
			return (1);
		}
	}
	return (-1);
}


int NewCache( void )
{
	register int  i;
	i = (cache_ptr + 1) % NUM_BUFS;
	if( BUFDIRTY(i))
		WriteIf( BUFHANDLE(i), BUFBLOCK(i).brec, (char *) &BUFBLOCK(i), sizeof(BLOCK));
	BUFHANDLE(i) = pci->ixfile;
	BUFDIRTY(i) = 0;
	return (i);
}


void LoadCache( RECPOS r )
{
	cache_ptr = NewCache();
	ReadIf( r, (char *)&BUFBLOCK( cache_ptr ), sizeof(BLOCK));
}


void GetCache( RECPOS r )
{
	if ( FindCache(r) < 0)
		LoadCache(r);
	block_ptr = &BUFBLOCK( cache_ptr );
}


void RetrieveBlock( int j, RECPOS r)
{
	if( j == 0)
		block_ptr = &(pci->root);
	else
		GetCache( r );
	CB(j) = block_ptr->brec;
}

//
// low level functions of BPLUS indexing routines
//

int PrevEntry( int off)
{
	if( off > 0 )
		off = ScanBlock( off );
	else
	{
		off = -1;
		CO(pci->level) = off;
	}
	return (off);
}


int NextEntry( int off )
{
	if( off == -1)
		off = 0;
	else
	{
		if( off < block_ptr->bend )
			off += ENT_SIZE;
	}
	CO(pci->level) = off;
	return (off);
}


void CopyEntry( ENTRY *to, ENTRY *from)
{
	memcpy( to, from, ENT_SIZE);
}


int ScanBlock( int n)
{
	register int off, last;
	off = 0;
	last = -1;
	while( off < n )
	{
		last = off;
		off += ENT_SIZE;
	}
	CO(pci->level) = last;
	return (last);
}

int LastEntry()
{
	return( ScanBlock( block_ptr->bend) );
}


//
//  maintain list of free index blocks
//

void WriteFree( RECPOS r, BLOCK *pb )
{
	pb->p0 = FREE_BLOCK;
	pb->brec = pci->dx.ff;
	WriteIf( pci->ixfile, r, (char*)pb, sizeof(BLOCK));
	pci->dx.ff = r;
}

RECPOS GetFree( void )
{
	RECPOS  r, rt;

	r = pci->dx.ff;
	if( r != NULLREC )
	{
		ReadIf( r, (char*)&rt, sizeof( RECPOS ));
		pci->dx.ff = rt;
	}
	else
#ifdef WIN32
		r = _lseek( pci->ixfile, 0L, SEEK_END );
#else
		r = lseek( pci->ixfile, 0L, SEEK_END );
#endif
	return (r);
}


//
//  general BPLUS block level functions indexing routines
//

int  FindBlock( ENTRY *pe, int *off, int *poff)
{
	register int pos, nextpos, ret;
	pos = -1;
	nextpos = 0;
	ret = 1;
	while( nextpos < block_ptr->bend)
	{
		if(pe->key <= ENT_ADR(block_ptr, nextpos)->key)
		{
			ret = (pe->key == ENT_ADR(block_ptr, nextpos)->key) ? 0 : -1;
			break;
		}

		pos = nextpos;
		nextpos += ENT_SIZE;
		/* nextpos = next_entry(pos); */
	}
	*poff = pos;
	if( ret == 0)
		*off = nextpos;
	else
		*off = pos;
	CO(pci->level) = *off;

	return (int)(ret);
}


void MoveDown( BLOCK *pb, int off, int n)
{
	memmove( ENT_ADR( pb, off), ENT_ADR( pb, off + n), pb -> bend - (off + n));
}


void MoveUp( BLOCK *pb, int off, int n)
{
    memmove( ENT_ADR( pb, off + n), ENT_ADR( pb, off), pb->bend - off);
}


void  InsertBlock( BLOCK *pb, ENTRY *pe, int off)
{
    MoveUp( pb, off, ENT_SIZE);
    CopyEntry( ENT_ADR( pb, off), pe);
    pb->bend += ENT_SIZE;
}


void  DeleteBlock( BLOCK *pb, int off)
{
	MoveDown( pb, off, ENT_SIZE);
	pb->bend -= ENT_SIZE;
}


//
// position at start/end of index routines
//

int FirstKey( ENTRY *pe, IX_DESC *pix)
{
	int ret;

	pci = pix;
	block_ptr = &(pci->root);
	CB(0) = 0L;
	CO(0) = -1;
	pci->level = 0;
	while( block_ptr->p0 != NULLREC)
	{
		RetrieveBlock( ++(pci->level), block_ptr->p0);
		CO(pci->level) = -1;
	}
	ret = NextKey( pe, pix);
	return (ret);
}


int LastKey( ENTRY *pe, IX_DESC *pix)
{
	long  ads;
	int   ret;

	pci = pix;
	block_ptr = &(pci->root);
	CB(0) = 0L;
	pci->level = 0;
	if( LastEntry() >= 0)
	{
		while( (ads = ENT_ADR( block_ptr, LastEntry())->idxptr) != NULLREC)
			RetrieveBlock( ++(pci->level), ads);
	}
	CO(pci->level) = block_ptr->bend;
	ret = PrevKey( pe, pix);
	return (ret);
}

//
// get next, previous entries
//

int NextKey( ENTRY *pe, IX_DESC *pix )
{
	RECPOS  address;
	pci = pix;
	RetrieveBlock( pci->level, CB(pci->level));
	if( CO( pci->level) == -1)
		address = block_ptr->p0;
	else
	{
		if (CO(pci->level) == block_ptr->bend)
			address = NULLREC;
		else
			address = ENT_ADR(block_ptr, CO(pci->level))->idxptr;
	}

	while( address != NULLREC)
	{
		RetrieveBlock( ++(pci->level), address);
		CO(pci->level) = -1;
		address = block_ptr->p0;
	}
	NextEntry( CO( pci->level));
	if( CO(pci->level) == block_ptr->bend)
	{
		do
		{
			if(pci->level == 0)
				return (EOIX);

			--(pci->level);
			RetrieveBlock(pci->level, CB(pci->level));
			NextEntry( CO( pci->level));
		}while( CO(pci->level) == block_ptr->bend);
	}
	CopyEntry( pe, ENT_ADR( block_ptr, CO(pci->level)));
	return (IX_OK);
}


int PrevKey( ENTRY *pe, IX_DESC *pix)
{
	RECPOS  address;
	pci = pix;
	RetrieveBlock( pci->level, CB(pci->level));
	PrevEntry( CO(pci->level));

	if( CO(pci->level) == -1)
		address = block_ptr->p0;
	else
		address = ENT_ADR(block_ptr, CO(pci->level))->idxptr;

	if( address != NULLREC)
	{
		do
		{
			RetrieveBlock( ++(pci->level), address);
			address = ENT_ADR( block_ptr, LastEntry())->idxptr;
		}while( address != NULLREC);
	}
	if( CO(pci->level) == -1)
	{
		do
		{
			if( pci->level == 0)
				return (EOIX);

			--(pci->level);
		}while( CO( pci->level) == -1);
		RetrieveBlock( pci->level, CB(pci->level));
	}
	CopyEntry( pe, ENT_ADR( block_ptr, CO( pci->level)));
	return ( IX_OK );
}

//
// insert new entries into tree
//

void Split( int l, ENTRY *pe, ENTRY *e )
{
	int  half, ins_pos;
	ins_pos = CO( pci->level);
	half = ScanBlock( (block_ptr->bend >> 1) + sizeof(RECPOS));
	if( half == ins_pos)
		*e = *pe;
	else
	{
		CopyEntry( e, ENT_ADR( block_ptr, half));
		MoveDown( block_ptr, half, ENT_SIZE);
		block_ptr->bend -= ENT_SIZE;
	}
	spare_block = &BUFBLOCK( NewCache());
	memmove( spare_block->entries, ENT_ADR(block_ptr,half), block_ptr->bend - half);
	spare_block->brec = GetFree();
	spare_block->bend = block_ptr->bend - half;
	spare_block->p0 = e->idxptr;
	block_ptr->bend = half;
	e->idxptr = spare_block->brec;

	if( ins_pos < half)
		InsertBlock( block_ptr, pe, ins_pos);
	else if( ins_pos > half)
	{
		ins_pos -= ENT_SIZE;
		InsertBlock( spare_block, pe, ins_pos - half);
		CB(l) = e->idxptr;
		CO(l) = CO(l) - half;
	}
	WriteIf( pci->ixfile, spare_block->brec, (char*)spare_block, sizeof(BLOCK));
}


void InsertLevel( int l, ENTRY *e )
{
	int  i;

	if( l < 0)
	{
		for( i = 1; i < MAX_LEVELS; i++)
		{
			CO( MAX_LEVELS - i) = CO( MAX_LEVELS - i - 1);
			CB( MAX_LEVELS - i) = CB( MAX_LEVELS - i - 1);
		}
		memmove( spare_block, &(pci->root), sizeof(BLOCK));
		spare_block->brec = GetFree();

		WriteIf( pci->ixfile, spare_block->brec, (char*)spare_block, sizeof(BLOCK));
		pci->root.p0 = spare_block->brec;

		CopyEntry( (ENTRY*)( pci->root.entries), e);
		pci->root.bend = ENT_SIZE;
		CO(0) = 0;
		pci->level = 0;
		(pci->dx.nl)++;
		pci->root_dirty = TRUE;
	}
	else
		InsertBlock( block_ptr, e, CO(l));
}


int InsertIx( ENTRY *pe, IX_DESC *pix)
{
	ENTRY    e, ee;
	int h;
	h = 0;
	pci = pix;
	ee = *pe;
//	do
	pci->dx.totalkeys++;	// increase the number of keys in the index
	pci->root_dirty = TRUE;
	for(;;)
	{
		if(CO(pci->level) < 0)
			CO(pci->level) = 0;
		else
			CO(pci->level) += ENT_SIZE;

		UpdateBlock();

		if( (block_ptr->bend + ENT_SIZE) > (unsigned int)split_size)
		{
			h = 1;
			Split( pci->level, &ee, &e);
			ee = e;
			pci->level--;
			if( pci->level < 0)
			{
				InsertLevel( pci->level, &e);
				break;
			}
			RetrieveBlock( pci->level, CB(pci->level));
		}
		else
		{
			InsertLevel( pci->level, &ee);
			break;
		}
	}
//	while (1);
	if( h )
		FindIx( pe, pix, 0);
	return (IX_OK);
}


//
// BPLUS find and add key functions
//

int FindIx( ENTRY *pe, IX_DESC *pix, int find)
{
	int      level, off, poff, ret;
	int      savelevel, saveoff, h;
	RECPOS   ads, saveads;
	pci = pix;
	ads = 0L;
	level = 0;
	h = 0;
	ret = 0;
	savelevel = 0;
	saveoff = 0;
	saveads = NULLREC;
	while( ads != NULLREC)
	{
		pci->level = level;
		RetrieveBlock( level, ads);
		ret = ret | !FindBlock( pe, &off, &poff);
		if( ret && find && !pci->duplicate )
			break;
		if( ret && pci->duplicate)
		{
			saveads = ads;
			savelevel = level;
			saveoff = off;
			off = poff;
			h = 1;
		}
		if( off == -1)
			ads = block_ptr->p0;
		else
			ads = ENT_ADR(block_ptr, off)->idxptr;
		CO(level++) = off;
	}
	if( h && find)
	{
		if (!ret)
		{
			RetrieveBlock( savelevel, saveads);
			pci->level = savelevel;
			ret = 1;
		}
		CO( savelevel) = saveoff;
	}
	return ( ret );
}

int FindKey( ENTRY *pe, IX_DESC *pix )
{
	int ret;

	pci = pix;
	if( (ret = FindIx( pe, pix, 1)) != 0)
		CopyEntry( pe, ENT_ADR( block_ptr, CO(pci->level)));
	return ( ret );
}


int AddKey( ENTRY *pe, IX_DESC *pix)
{
	int ret;
	ret = FindIx( pe, pix, 0);
	if( ret && (pci->duplicate == 0))
		return (IX_FAIL);
	pe->idxptr = NULLREC;
	return (InsertIx( pe, pix));
}


int LocateKey( ENTRY *pe, IX_DESC *pix )
{
	int ret;
	ret = FindIx( pe, pix, 1);
	if( ret )
		CopyEntry( pe, ENT_ADR( block_ptr, CO(pci->level)));
	else if( NextKey( pe, pix) == EOIX)
		ret = EOIX;
	return ( ret );
}

int FindExact( ENTRY *pe, IX_DESC *pix)
{
	int  ret;
	ENTRY e;

	pci = pix;
	CopyEntry( &e, pe);
	ret = FindKey(&e, pix);
	if( ret && pix->duplicate && e.key != pe->key)
	{
		while( e.recptr != pe->recptr)
		{
			if( NextKey( &e, pci) == EOIX)
				return (0);   // no more records

			if( e.key == pe->key )
				break;	  	// key changed
		}
	}
	CopyEntry( pe, &e);
	return (ret);
}

//
// BPLUS delete key functions
//

int DeleteKey( ENTRY *pe, IX_DESC *pix)
{
	ENTRY   e;
	RECPOS  ads;
	int     h, leveli, levelf;

	pci = pix;

	if( !FindExact( pe, pix))
		return( IX_FAIL );
//	h = 1;
	pix->dx.totalkeys--; // decrease the number of keys in the index
	pci->root_dirty = TRUE;
	if( (ads = pe->idxptr) != NULLREC)
	{
		leveli = pci->level;
		do
		{
			RetrieveBlock( ++(pci->level), ads);
			CO(pci->level) = -1;
		}
		while( (ads = block_ptr->p0) != NULLREC);
		CO(pci->level) = 0;
		CopyEntry( &e, ENT_ADR( block_ptr, CO(pci->level)));
		levelf = pci->level;
		pci->level = leveli;
		ReplaceEntry( &e);
		pci->level = levelf;
	}

	//while( h )
	do
	{
		RetrieveBlock( pci->level, CB( pci->level));
		DeleteBlock( block_ptr, CO( pci->level));
		UpdateBlock();
		if( (pci->level == 0) && (block_ptr->bend == 0))
		{
			// tree was reduced in height
			if( pci->root.p0 == NULLREC)
				break;

			RetrieveBlock( ++pci->level, pci->root.p0);
			memmove( &(pci->root), block_ptr, sizeof(BLOCK));
			(pci->dx.nl)--;
			WriteFree( block_ptr->brec, block_ptr);
			BUFDIRTY( cache_ptr ) = 0;
			BUFHANDLE( cache_ptr ) = 0;
		}
		h = (block_ptr->bend < comb_size) && (pci->level > 0);
		if( h )
			h = CombineBlock( CB( pci->level), block_ptr->bend);
	}while( h );
	FindIx( pe, pix, 0);
	return(IX_OK);
}


int CombineBlock( RECPOS ads, int size)
{
	ENTRY  e;
	RECPOS address;
	int    off, ret, saveoff, ibuff;
	ret = 0;
	saveoff = CO( --(pci->level));
	RetrieveBlock( pci->level, CB(pci->level));
	if( (off = NextEntry( saveoff )) < block_ptr->bend)
	{
		// combine with page on right
		if( (size + ENT_SIZE) < (unsigned int)split_size)
		{
			CopyEntry( &e, ENT_ADR( block_ptr, off));
			address = ENT_ADR( block_ptr, CO( pci->level))->idxptr;
			RetrieveBlock( ++pci->level, address);
			ibuff = cache_ptr;
			spare_block = block_ptr;
			RetrieveBlock( pci->level, ads);
			
			if(((block_ptr->bend + spare_block->bend + ENT_SIZE) >= split_size) && (spare_block->bend <= block_ptr->bend + ENT_SIZE))
				return( ret );

			e.idxptr = spare_block->p0;
			InsertBlock( block_ptr, &e, block_ptr->bend);
			UpdateBlock();
			if( (block_ptr->bend + spare_block->bend) < split_size)
			{
				// combine the blocks
				memmove(ENT_ADR(block_ptr, block_ptr->bend), ENT_ADR( spare_block, 0), spare_block->bend);
				block_ptr->bend += spare_block->bend;

				WriteFree( spare_block->brec, spare_block);
				BUFDIRTY(ibuff) = 0;
				BUFHANDLE(ibuff) = 0;
				--pci->level;
				ret = 1;
			}
			else
			{
				// move an entry up to replace the one moved
				CopyEntry( &e, ENT_ADR( spare_block, 0));
				MoveDown( spare_block, 0, ENT_SIZE);
				spare_block->bend -= ENT_SIZE;
				spare_block->p0 = e.idxptr;
				BUFDIRTY(ibuff) = 1;
				--(pci->level);
				ReplaceEntry(&e);
			}
		}
	}
	else
	{
		// move from page on left
		if( size + ENT_SIZE < (unsigned int)split_size)
		{
			CopyEntry( &e, ENT_ADR( block_ptr, saveoff));
			off = PrevEntry( saveoff);
			if( CO(pci->level) == -1)
				address = block_ptr->p0;
			else
				address = ENT_ADR( block_ptr, CO(pci->level))->idxptr;

			RetrieveBlock( ++pci->level, address);
			off = LastEntry();
			ibuff = cache_ptr;
			spare_block = block_ptr;
			RetrieveBlock( pci->level, ads);
			if( ((block_ptr->bend + spare_block->bend + ENT_SIZE) >= split_size) && (spare_block->bend <= block_ptr->bend + ENT_SIZE))
				return( ret );
			BUFDIRTY(ibuff) = 1;
			CO(pci->level) = 0;
			e.idxptr = block_ptr->p0;
			InsertBlock( block_ptr, &e, 0);
			if( (block_ptr->bend + spare_block->bend) < split_size)
			{
				// combine the blocks
				memmove(ENT_ADR(spare_block, spare_block->bend), ENT_ADR( block_ptr, 0), block_ptr->bend);
				spare_block->bend += block_ptr->bend;

				WriteFree( block_ptr->brec, block_ptr);
				BUFDIRTY( cache_ptr ) = 0;
				BUFHANDLE( cache_ptr ) = 0;
				CO(--(pci->level)) = saveoff;
				ret = 1;
			}
			else
			{
				// move an entry up to replace the one moved
				block_ptr->p0 = ENT_ADR( spare_block,off)->idxptr;

				CopyEntry( &e, ENT_ADR( spare_block, off));
				spare_block->bend = off;
				UpdateBlock();
				CO(--(pci->level)) = saveoff;
				ReplaceEntry( &e);
			}
		}
	}
	return (ret);
}


void ReplaceEntry( ENTRY *pe )
{
	RetrieveBlock( pci->level, CB(pci->level));
	pe->idxptr = ENT_ADR( block_ptr, CO( pci->level))->idxptr;
	DeleteBlock( block_ptr, CO( pci->level));
	PrevEntry( CO(pci->level));
	InsertIx( pe, pci);
	UpdateBlock();
}

int  GetTotalKeys( IX_DESC *pix )
{
	return (pix->dx.totalkeys);
}


