//#include "stdafx.h"

#include <stdio.h>
#include <fcntl.h>
#include <unistd.h>

#if OPH1004 || OPN2002
#include <fcntl.h>
#include <sys\types.h>            // delete this line for Turbo C
#include <sys\stat.h>
#endif

#include <string.h>


//#include "Database.h"
#include "DBaseFix.h"
#include "BPLUS.H"

#if defined(WIN32)
	#include <io.h>
#endif
#include "lib.h"


#if OPH1004 || OPN2002
#include <unistd.h>
#undef O_BINARY
#define O_BINARY 0x00
#endif

#ifdef WIN32
	#define open		_open
	#define close		_close
	#define read		_read
	#define write		_write
	#define lseek		_lseek
	#define O_RDWR		_O_RDWR 
	#define O_BINARY	_O_BINARY
	#define O_CREAT		_O_CREAT 
	#define O_TRUNC		_O_TRUNC  

long fsize(const char *FileName)
{
    struct stat file;
     
     if(!stat(FileName,&file))
     {
         return file.st_size;
     }

   
     return 0;
}

#endif

extern char g_recordBuffer[];
extern char g_fieldBuffer[];

int FixGetField( char *fieldBuffer, const char *buffer, int offset, int fieldLength )
{
	if( buffer == NULL || buffer[0] == '\0')
		return -1;
	strncpy( fieldBuffer, buffer + offset, fieldLength );
	fieldBuffer[ fieldLength ] = '\0';
	return fieldLength;
}

int FixBuildIndex( const char *dbName, const char *idxName, int offset, int maxFieldLength, int dup )
{
	static IX_DESC ix;
	static ENTRY e;
	FILE *fp;
	int error;
	long filePos = 0L;
	int fieldLength;
	//static char fieldBuffer[ MAX_VAR_SIZE + 1];
	// Open the database in binary read only mode

	if( (fp = fopen( dbName, "rb")) == NULL )
		return ERR_OPEN_DATABASE;

	if( MakeIndex( idxName, &ix, dup ) == IX_FAIL )
	{
		fclose( fp );
		return ERR_OPEN_INDEX; 
	}
	error = OK;

	while( fgets( g_recordBuffer, MAX_G_BUFFER, fp ) != NULL )
	{
		// Fixed record string
		if( (fieldLength = FixGetField( g_fieldBuffer, g_recordBuffer, offset, maxFieldLength  )) < 0 )
		{
			error = fieldLength; 
			goto build_index_error;
		}
		
		e.key = hash(g_fieldBuffer);
		e.recptr  = (RECPOS)filePos;

		// Add the key to the indexing file
		if( AddKey( &e, &ix ) != IX_OK )
		{
			error = ERR_INDEX_DUPLICATE;
			goto build_index_error;
		}
		filePos = ftell( fp );
	}

build_index_error:
	fclose( fp );
	CloseIndex( &ix );
	return error;
}

int FixRebuildIndex(  SDatabaseItem *dbItem, SDBOutVal *dbOut )
{
	int ret;
	CloseDatabase( dbOut );

	if( (ret = FixBuildIndex( dbItem->dbFilename, dbItem->idxFilename, dbItem->field[ dbItem->keyfield ].offsetField, dbItem->field[ dbItem->keyfield ].maxField, dbItem->flags & F_DB_APPEND )) < OK )
	{
		// Error while creating index file, so remove it
		remove( dbItem->idxFilename );
		return ret;	// ERROR!!!!
	}

	if( OpenIndex( dbItem->idxFilename, &(dbOut->ix), dbItem->flags & F_DB_APPEND ) != IX_OK )
		return ERR_OPEN_INDEX;

	dbOut->fdDb = open( dbItem->dbFilename, O_RDWR | O_BINARY, 0777 );
	return DATABASE_OK;
}


int FixOpenDatabase( SDatabaseItem *dbItem, SDBOutVal *dbOut )
{
	int ret;
	long lookupSize = 0L;

	if( (lookupSize = fsize( dbItem->dbFilename )) == -1L )
		return ERR_OPEN_DATABASE;

	if( (lookupSize % dbItem->maxRecordSize) != 0 )
		return ERR_DB_WRONG_FORMAT;

	dbOut->fdDb = open( dbItem->dbFilename, O_RDWR | O_BINARY, 0777 );

	if( dbOut->fdDb < 0 )
		return ERR_OPEN_DATABASE;

	if( OpenIndex( dbItem->idxFilename, &(dbOut->ix), dbItem->flags & F_DB_APPEND ) != IX_OK )
	{
		// Close the original database first
		close( dbOut->fdDb );

		if( (ret = FixBuildIndex( dbItem->dbFilename, dbItem->idxFilename, dbItem->field[ dbItem->keyfield ].offsetField, dbItem->field[ dbItem->keyfield ].maxField, dbItem->flags & F_DB_APPEND )) < OK )
		{
			// Error while creating index file, so remove it
			remove( dbItem->idxFilename );
			return ret;	// ERROR!!!!
		}	

		if( OpenIndex( dbItem->idxFilename, &(dbOut->ix), dbItem->flags & F_DB_APPEND ) != IX_OK )
			return ERR_OPEN_INDEX;

		dbOut->fdDb = open( dbItem->dbFilename, O_RDWR | O_BINARY, 0777 );
	}
	dbOut->dbFlags = dbItem->flags;
	return DATABASE_OK;
}

int FixCreateDatabase( SDatabaseItem *dbItem, SDBOutVal *dbOut )
{
	dbOut->fdDb = open( dbItem->dbFilename, O_RDWR | O_CREAT | O_TRUNC | O_BINARY, 0777 );

	if( dbOut->fdDb < 0 )
		return ERR_OPEN_DATABASE;

	if( MakeIndex( dbItem->idxFilename, &(dbOut->ix), dbItem->flags & F_DB_APPEND ) == IX_OK )
	{
		dbOut->dbFlags = dbItem->flags;
		return DATABASE_OK;
	}
	close( dbOut->fdDb );

	return ERR_OPEN_INDEX;	// Error creating the index file 
}


