//
// Database.c(pp)
//
// Implementation Database functions that finally call the fix or the csv database functions
// 
// Martin Jansen
// 
//#include "stdafx.h"
#include "database.h"
#include "dbasecsv.h"
#include "dbasefix.h"
#include <setjmp.h>
#include <string.h>
#include <stdio.h>
#include <fcntl.h>
#include <unistd.h>

#if defined(WIN32)
#include <io.h>
#endif
#include "lib.h"

#if OPH1004 || OPN2002
#include <unistd.h>
#undef O_BINARY
#define O_BINARY 0x00
#endif

#ifdef WIN32
#define open		_open
#define close		_close
#define read		_read
#define write		_write
#define lseek		_lseek
#define tell		_tell
#define chsize		_chsize
#define O_RDWR		_O_RDWR
#define O_BINARY	_O_BINARY
#define O_CREAT		_O_CREAT
#define O_TRUNC		_O_TRUNC
#endif

//#define MAX_FIELD_SIZE	128
//#define MAX_G_BUFFER	1024
//#define MAX_FIELD		50

extern jmp_buf catchIndexError;		// environment storage.

//int g_maxRecordSize = 0;

char g_recordBuffer[MAX_G_BUFFER + 1];
char g_fieldBuffer[MAX_G_FIELD + 1];

static int AreDatabasesOpen(SDBOutVal *dbOutVal);

//static void RemoveCrLf( char *outputBuffer );

int GetField(const SDatabaseItem *dbItem, char *fieldBuffer, const char *buffer, int fieldNr, int maxFieldSize) 
{
	if (dbItem->flags & F_DB_CSV)
		return CsvGetField(fieldBuffer, buffer, fieldNr, maxFieldSize, dbItem->csvChar);
	else
		return FixGetField(fieldBuffer, buffer, dbItem->field[fieldNr].offsetField, dbItem->field[fieldNr].maxField);
}

int OpenDatabase(const SDatabaseItem *dbItem, SDBOutVal *dbOut) 
{
	int jmpError;
	if ((jmpError = setjmp(catchIndexError)) != 0) 
	{
		return jmpError;
	}

//	g_maxRecordSize = dbItem->maxRecordSize;
	if (dbItem->flags & F_DB_CSV)
		return CsvOpenDatabase((SDatabaseItem *) dbItem, dbOut);
	else
		return FixOpenDatabase((SDatabaseItem *) dbItem, dbOut);
}

int CreateDatabase(const SDatabaseItem *dbItem, SDBOutVal *dbOut) 
{
	int jmpError;
	if ((jmpError = setjmp(catchIndexError)) != 0) {
		return jmpError;
	}

	//g_maxRecordSize = dbItem->maxRecordSize;
	if (dbItem->flags & F_DB_CSV)
		return CsvCreateDatabase((SDatabaseItem *) dbItem, dbOut);
	else
		return FixCreateDatabase((SDatabaseItem *) dbItem, dbOut);
}

void CloseDatabase(SDBOutVal *dbOut) 
{
	if (setjmp(catchIndexError) != 0)
		return; // Just return

	// Close the index file
	CloseIndex(&(dbOut->ix));

	// Close the 
	close(dbOut->fdDb);

	dbOut->fdDb = -1;
}

static int AreDatabasesOpen(SDBOutVal *dbOut) 
{
	if (dbOut->fdDb < 0)
		return ERR_OPEN_DATABASE;
	if (dbOut->ix.ixfile < 0)
		return ERR_OPEN_INDEX;

	return DATABASE_OK;
}

int CutOffAfterCrLf(char *outputBuffer) 
{
	char *ptr = NULL;
	char *ptr2 = NULL;

	ptr = strchr(outputBuffer, '\r');
	ptr2 = strchr(outputBuffer, '\n');

	// Hack: Bugfix to solve crash when files contains both '\n' and '\r\n' line endings
	if (ptr2 != NULL && ptr != NULL && ptr2 < ptr) {
		ptr2++;
		*ptr2 = '\0';
		return ptr2 - outputBuffer;
	}

	// Cut off the first '\r' '\n'
	if (ptr != NULL) {
		ptr++;
		if (*ptr == '\n')
			ptr++;
		*ptr = '\0';
		return ptr - outputBuffer;
	}

	// Check for only '\n'
	if (ptr2 != NULL) {
		ptr2++;
		*ptr2 = '\0';
	}

	return ptr2 - outputBuffer;
}

//int ReadRecordFromDB( int fd, long recptr, char *outputBuffer )
int ReadRecord(const SDatabaseItem *dbItem, SDBOutVal *dbOut, long recptr, char *outputBuffer) 
{
	if (outputBuffer == NULL)	// hack
		return DATABASE_OK;

	if (lseek(dbOut->fdDb, recptr, SEEK_SET) == -1L)
		return ERR_INDEX_KEY_NF; // TODO: maybe change this in some other value

	if (read(dbOut->fdDb, outputBuffer, dbItem->maxRecordSize) <= 0)
		return ERR_INDEX_KEY_NF; // TODO: maybe change this in some other value

	CutOffAfterCrLf(outputBuffer);
	return DATABASE_OK;
}

int SearchDatabase(const SDatabaseItem *dbItem, SDBOutVal *dbOut, const char *searchKey, long *precPtr, char *outputBuffer) 
{
	ENTRY e;
	int error;
	*precPtr = -1L;

	if ((error = AreDatabasesOpen(dbOut)) != DATABASE_OK)
		return ERROR;
	if ((error = setjmp(catchIndexError)) != 0)
		return error;

	e.key = hash(searchKey);

	dbOut->ix.duplicate = (dbItem->flags & F_DB_APPEND) ? TRUE : FALSE;
		
	if (FindKey(&e, &(dbOut->ix)) == 0)
		return ERR_INDEX_KEY_NF;

	*precPtr = e.recptr;
	return ReadRecord(dbItem, dbOut, e.recptr, outputBuffer);
}

int ReadFirstRecord(const SDatabaseItem *dbItem, SDBOutVal *dbOut, long *precptr, char *outputBuffer) 
{
	ENTRY e;
	int error;

	if ((error = AreDatabasesOpen(dbOut)) != DATABASE_OK)
		return ERROR;
	if ((error = setjmp(catchIndexError)) != 0)
		return error;

	if ((error = FirstKey(&e, &(dbOut->ix))) != IX_OK)
		return error;

	*precptr = e.recptr;
	return ReadRecord(dbItem, dbOut, e.recptr, outputBuffer);
}

int ReadLastRecord(const SDatabaseItem *dbItem, SDBOutVal *dbOut, long *precptr, char *outputBuffer) 
{
	ENTRY e;
	int error;

	if ((error = AreDatabasesOpen(dbOut)) != DATABASE_OK)
		return ERROR;
	if ((error = setjmp(catchIndexError)) != 0)
		return error;

	if ((error = LastKey(&e, &(dbOut->ix))) != IX_OK)
		return error;

	*precptr = e.recptr;
	return ReadRecord(dbItem, dbOut, e.recptr, outputBuffer);
}

int ReadPreviousRecord(const SDatabaseItem *dbItem, SDBOutVal *dbOut, long *precptr, char *outputBuffer) 
{
	ENTRY e;
	int error;

	if ((error = AreDatabasesOpen(dbOut)) != DATABASE_OK)
		return ERROR;
	if ((error = setjmp(catchIndexError)) != 0)
		return error;

	if ((error = PrevKey(&e, &(dbOut->ix))) != IX_OK)
		return error;

	*precptr = e.recptr;
	return ReadRecord(dbItem, dbOut, e.recptr, outputBuffer);
}

int ReadNextRecord(const SDatabaseItem *dbItem, SDBOutVal *dbOut, long *precptr, char *outputBuffer) 
{
	ENTRY e;
	int error;

	if ((error = AreDatabasesOpen(dbOut)) != DATABASE_OK)
		return ERROR;
	if ((error = setjmp(catchIndexError)) != 0)
		return error;

	if ((error = NextKey(&e, &(dbOut->ix))) != IX_OK)
		return error;

	*precptr = e.recptr;
	return ReadRecord(dbItem, dbOut, e.recptr, outputBuffer);
}

// adapted from String.hashCode()
unsigned long hash(const char *str)		// 32-bit hash key
{
    unsigned long hash = 5381;
    int c;

    while ( (c = *str++) != 0)
        hash = ((hash << 5) + hash) + c; /* hash * 33 + c */

    return hash;
}

//long long hash(const char *str)		// 64-bit hash key
//{
//	long long h = 1125899906842597; // prime
//	
//	int i, len = strlen(str);
//
//	for (i=0; i<len; i++) 
//	{
//		h = 31*h + str[i];
//	}
//
//	return h;
//}

int WriteRecord(const SDatabaseItem *dbItem, SDBOutVal *dbOut, long *precPtr, char *record, int flag) 
{
	int error;
	int length;
	long fileSize = 1L;
	long copyFrom;
	int copyData;
	ENTRY e;

	if ((error = AreDatabasesOpen(dbOut)) != DATABASE_OK)
		return ERROR;
	if ((error = setjmp(catchIndexError)) != 0)
		return error;

	length = strlen(record);
	if ((*precPtr == -1L)) // || (dbItem->flags & F_DB_APPEND)
	{
		// Append to the end of the database
		*precPtr = lseek(dbOut->fdDb, 0L, SEEK_END);
		if (write(dbOut->fdDb, record, length) != length)
			return ERR_DB_WRITE;
		GetField(dbItem, g_fieldBuffer, record, dbItem->keyfield, MAX_G_FIELD);
		
		e.key = hash(g_fieldBuffer);
		e.recptr = *precPtr;
		AddKey(&e, &(dbOut->ix));
	} 
	else 
	{
		// Update the record
		if ((error = ReadRecord(dbItem, dbOut, *precPtr, g_recordBuffer)) != DATABASE_OK)
			return error;
		if (strlen(g_recordBuffer) == (size_t) length) {
			// Overwrite the current record
			lseek(dbOut->fdDb, *precPtr, SEEK_SET);
			if (write(dbOut->fdDb, record, length) != length)
				return ERR_DB_WRITE;
			return DATABASE_OK;
		}
		// This should be a csv file, copy all data over the record
		fileSize = fsize(dbItem->dbFilename);

		copyFrom = (*precPtr) + strlen(g_recordBuffer);

		while (copyFrom < fileSize) {
			lseek(dbOut->fdDb, copyFrom, SEEK_SET);
			if ((copyData = read(dbOut->fdDb, g_recordBuffer, MAX_G_BUFFER))
					< 1)
				return ERR_DB_READ;
			copyFrom = tell(dbOut->fdDb);
			lseek(dbOut->fdDb, *precPtr, SEEK_SET);
			if (write(dbOut->fdDb, g_recordBuffer, copyData) != copyData)
				return ERR_DB_WRITE;

			*precPtr = tell(dbOut->fdDb);
		}
		chsize(dbOut->fdDb, *precPtr);
		lseek(dbOut->fdDb, 0, SEEK_END);// Hack: fixes bug that occurs when copyFrom == filesize

		if (write(dbOut->fdDb, record, length) != length)
			return ERR_DB_WRITE;

		// create a totally new index file
		return CsvRebuildIndex((SDatabaseItem *) dbItem, dbOut);
	}
	return DATABASE_OK;
}

int DeleteRecord(const SDatabaseItem *dbItem, SDBOutVal *dbOut, long recptr) 
{
	int error;
	long fileSize;
	long copyFrom;
	int copyData;

	if ((error = ReadRecord(dbItem, dbOut, recptr, g_recordBuffer))	!= DATABASE_OK)
		return error;

	fileSize = fsize(dbItem->dbFilename);
	copyFrom = recptr + strlen(g_recordBuffer);
	while (copyFrom < fileSize) 
	{
		lseek(dbOut->fdDb, copyFrom, SEEK_SET);
		if ((copyData = read(dbOut->fdDb, g_recordBuffer, MAX_G_BUFFER)) < 1)
			return ERR_DB_READ;
		copyFrom = tell(dbOut->fdDb);
		lseek(dbOut->fdDb, recptr, SEEK_SET);
		if (write(dbOut->fdDb, g_recordBuffer, copyData) != copyData)
			return ERR_DB_WRITE;
		recptr = tell(dbOut->fdDb);
	}
	chsize(dbOut->fdDb, recptr);

	// create a totally new index file
	if (dbItem->flags & F_DB_CSV)
		return CsvRebuildIndex((SDatabaseItem *) dbItem, dbOut);
	else
		return FixRebuildIndex((SDatabaseItem *) dbItem, dbOut);
}
