//	BPLUS.H
//
//	Version: 1.3a
//
//	---------------------------------------------------------------------------
//	BPLUS file indexing routines
//			by
//	Youri Tseytins.
//	---------------------------------------------------------------------------
//
//	Copyright (c) 2005
//
//  This software is provided 'as-is', without any express or implied
//  warranty.  In no event will the authors be held liable for any damages
//  arising from the use of this software.
//
//  Permission is granted to anyone to use this software for any purpose,
//  including commercial applications, and to alter it and redistribute it
//  freely, subject to the following restrictions:
//
//  1. The origin of this software must not be misrepresented; you must not
//     claim that you wrote the original software. If you use this software
//     in a product, an acknowledgment in the product documentation would be
//     appreciated but is not required.
//  2. Altered source versions must be plainly marked as such, and must not be
//     misrepresented as being the original software.
//  3. This notice may not be removed or altered from any source distribution.
//
//

// Sources changed to make it work on Opticon Terminals
// Martin Jansen

#ifndef ___BPLUS__H___
#define ___BPLUS__H___

#if defined(__cplusplus)
#define ANSI_C
extern "C"{
#endif
#include <setjmp.h>

#define IX_OK       1
#define IX_FAIL     0
#define LOCK_FAIL  (-1)
#define EOIX       (-2)
#define IX_READ		(-3)
#define IX_WRITE	(-4)

//#define MAXKEY      (50+2)		// 100
#define NUM_BUFS    8
#define MAX_LEVELS  32
#define IXB_SIZE    512
#define IXB_SPACE  (IXB_SIZE - sizeof(int) - (sizeof(long) << 1))

typedef long RECPOS;

typedef unsigned long HASHKEY;

typedef struct						// entry structure in index       
{  
	RECPOS   idxptr;				// points to lower index level     
    RECPOS   recptr;				// points to data record 
    HASHKEY key;					// hash key             
} ENTRY;

typedef struct						// index record format
{  
	RECPOS	brec;					// position in index file or location of next free block  
	int     bend;					// first unused block location     
	RECPOS	p0;						// points to next level            
	char	entries[IXB_SPACE];		// here are the key entries 
} BLOCK;

typedef struct						// disk file info                  
{  
	RECPOS	ff;						// location of first free block
    int		nl;						// number of index levels
	int		totalkeys;				// number of keys in the index	TEST TEST TEST
} IX_DISK;

typedef struct						// memory buffer pool of indx blks 
{  
	int		dirty;					// true if changed  
    int		handle;					// index file handle               
	int     count;					// number of times referenced      
    BLOCK	mb;
} MEMBLOCK;

typedef struct
{  
	MEMBLOCK	cache [ NUM_BUFS ];
} IX_BUFFER;

typedef struct						// in-memory index descriptor 
{  
	int		ixfile;					// file descriptor
    int		root_dirty;				// true if root changed
    int		level;					// level in btree
    int		duplicate;				// no duplicate keys if 0        
    struct
    {  
		RECPOS	cblock;				// position in index file 
		int		coffset;			// current offset within block
	} pos [ MAX_LEVELS ];
	BLOCK	root;					// root index record
	IX_DISK	dx;
} IX_DESC;

void BplusInit(void);
int  OpenIndex( const char *idxName, IX_DESC *pix, int dup);
int  MakeIndex( const char *idxName, IX_DESC *pix, int dup);
int  CloseIndex( IX_DESC *pix);

int  GetTotalKeys( IX_DESC *pix );

int  FirstKey( ENTRY *pe, IX_DESC *pix);
int  LastKey( ENTRY *pe, IX_DESC *pix);
int  NextKey( ENTRY *pe, IX_DESC *pix);
int  PrevKey( ENTRY *pe, IX_DESC *pix);

int  FindKey( ENTRY *pe, IX_DESC *pix);
int  AddKey( ENTRY *pe, IX_DESC *pix);
int  LocateKey( ENTRY *pe, IX_DESC *pix);
int  DeleteKey( ENTRY *pe, IX_DESC *pix);
int  FindExact( ENTRY *pe, IX_DESC *pix);
void BufferFlush( IX_DESC *pix);

//int	 BuildIndexCsv( const char *dbName, const char *idxName, int dup );
// TODO: add this function later
//int	 BuildIndexFixed( const char *name_db, const char *name_idx, unsigned char dup );


#if defined(__cplusplus)
}
#endif

#endif // ___BPLUS__H___

