//
// Database.h
//
// Database functions that finally call the fix or the csv functions
//
// Martin Jansen
//

#ifndef __SPEEDGEN_DBASE_H__
#define __SPEEDGEN_DBASE_H__

// include is needed to get the IX_DESC value
#include "BPLUS.h"

//
// Error code defines, these should match the fix database error code
//
#define DATABASE_OK			(0)			// OK message
#define ERR_OPEN_DATABASE	(-1)		// Cannot open the database file
#define ERR_OPEN_INDEX		(-2)		// cannot open/create the index file

#define ERR_INDEX_DUPLICATE	(-5)		// Duplicate key found in index while not allowed
#define ERR_INDEX_KEY_NF	(-6)		// Key has not been found.

#define ERR_DB_READ			(-3)
#define ERR_DB_WRITE		(-4)
#define ERR_INDEX_READ		(-10)		// Cannot read from index file
#define ERR_INDEX_WRITE		(-11)		// Cannot write to index file

#define ERR_DB_WRONG_FORMAT (-12)		// Fix size database has the wrong size

#define MAX_FIELDS		10

#define MAX_G_BUFFER    256
#define MAX_G_FIELD     80


//
// Special Database flags
//
enum
{
	F_DB_CSV	= 1,	// if this flag is set then it is a CSV style database
	F_DB_LOOKUP	= 2,	// if this flag is set then index file that cannot contain duplicates
	F_DB_DATE	= 4,	// the datafile database contains a date field
	F_DB_TIME	= 8,	// the datafile database contains a time field
	F_DB_APPEND = 16	// Append database multiple keys are allowed in the index files
};

typedef struct
{
	short offsetField;					// offset or the field number
	short maxField;						// max field size (csv) or field size fixed size db
}SFields;


typedef struct
{
	char *dbFilename;					// Database filename
	char *idxFilename;					// index filename
	unsigned char flags;				// Special options flags
	int	 maxRecordSize;					// The maximum record size including the <CR><LF> and CSV separator
	int  keyfield;						// The field number that holds the search key
	int  csvChar;						// the csv separator character
	SFields field[ MAX_FIELDS ];	    // The field description that makeup a record
}SDatabaseItem;


typedef struct
{
	int		fdDb;						// database file descriptor
	IX_DESC ix;							// holds the indexing values
	int		dbFlags;					// Special options flags
}SDBOutVal;

int OpenDatabase( const SDatabaseItem *dbItem, SDBOutVal *dbOut );
int CreateDatabase( const SDatabaseItem *dbItem, SDBOutVal *dbOut );
void CloseDatabase( SDBOutVal *dbOut );

int CsvBuildIndex( const char *dbName, const char *idxName, int delimiter, int fieldNr, int dup );

HASHKEY hash(const char *str);

int ReadFirstRecord( const SDatabaseItem *dbItem, SDBOutVal *dbOut, long *precptr, char *outputBuffer );
int ReadLastRecord( const SDatabaseItem *dbItem, SDBOutVal *dbOut, long *precptr, char *outputBuffer );
int ReadPreviousRecord( const SDatabaseItem *dbItem, SDBOutVal *dbOut, long *precptr, char *outputBuffer  );
int ReadNextRecord( const SDatabaseItem *dbItem, SDBOutVal *dbOut, long *precptr, char *outputBuffer );

int WriteRecord( const SDatabaseItem *dbItem, SDBOutVal *dbOut, long *precptr, char *record, int flag );
int ReadRecord( const SDatabaseItem *dbItem, SDBOutVal *dbOut, long recptr, char *outputBuffer );
int DeleteRecord( const SDatabaseItem *dbItem, SDBOutVal *dbOut, long recptr );

int GetField( const SDatabaseItem *dbItem, char *fieldBuffer, const char *buffer, int fieldNr, int maxFieldSize );
int SearchDatabase( const SDatabaseItem *dbItem, SDBOutVal *dbOut, const char *searchKey, long *precPtr, char *outputBuffer );

int CutOffAfterCrLf( char *outputBuffer );

#endif // __SPEEDGEN_DBASE_H__


